# -*- coding: ascii -*-
#
#  Copyright (C) 2003-2004 by Shyouzou Sugitani <shy@users.sourceforge.jp>
#  Copyright (C) 2003 by Shun-ichi TAHARA <jado@flowernet.gr.jp>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License (version 2) as
#  published by the Free Software Foundation.  It is distributed in the
#  hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the GNU General Public License for more details.
#

import os
import re
import types

if os.environ.has_key("DISPLAY"):
    import gtk

import ninix.pix
import ninix.sakura
import ninix.surface

class Menu:
    def __init__(self, ghost, surface, accelgroup):
        self.__ghost = ghost
        self.__surface = surface
        self.__menu_list = {
            'portal':    [_('Portal sites(_P)'), None, 1, 1, ('/portal', None, None, 0, '<Branch>')],
            'recommend': [_('Recommend sites(_R)'), None, 1, 1, ('/recommend', None, None, 0, '<Branch>')],
            'file':      [_('File(_F)'), None, 1, 1, ('/file', None, None, 0, '<Branch>')],
            'update':    [_('Network Update(_U)'), None, 1, 1, ('/file/update', None, self.__ghost.network_update, 0, '')],
            'vanish':    [_('Vanish(_F)'), None, 1, 1, ('/file/vanish', None, self.__ghost.vanish, 0, '')],
            'reload':    [_('Reload(_L)'), None, 1, 1, ('/file/reload', None, self.__ghost.reload, 0, '')],
            'settings':  [_('Settings...(_O)'), None, 1, 1, ('/file/settings', None, self.__ghost.edit_preferences_cb, 0, '')],
            'usage':     [_('Usage graph(_A)'), None, 1, 1, ('/file/usage', None, self.__ghost.show_usage, 0, '')],
            'version':   [_('Version(_V)'), None, 1, 1, ('/file/version', None, self.__ghost.about, 0, '')],
            'exit':      [_('Exit(_Q)'), None, 1, 1, ('/file/exit', None, self.__ghost.close, 0, '')],
            'change':     [_('Change(_G)'), None, 1, 1, ('/change', None, None, 0, '<Branch>')],
            'summon':     [_('Summon(_X)'), None, 1, 1, ('/summon', None, None, 0, '<Branch>')],
            'shell':     [_('Shell(_S)'), None, 1, 1, ('/shell', None, None, 0, '<Branch>')],
            'balloon':   [_('Balloon(_B)'), None, 1, 1, ('/balloon', None, None, 0, '<Branch>')],
            'scaling':   [_('Surface Scaling(_S)'), None, 1, 1, ('/scaling', None, None, 0, '<Branch>')],
            'costume':   [_('Costume(_C)'), None, 1, 1, ('/costume', None, None, 0, '<Branch>')],
            'stick':     [_('Stick(_Y)'), None, 1, 1, ('/stick', None, self.__surface.window_stick, 0, '<CheckItem>')],
            'wait':      [_('Script Wait(_W)'), None, 1, 1, ('/wait', None, None, 0, '<Branch>')],
            'nekodorif':    [_('Nekodorif(_N)'), None, 1, 1, ('/nekodorif', None, None, 0, '<Branch>')],
            'kinoko':      [_('Kinoko(_K)'), None, 1, 1, ('/kinoko', None, None, 0, '<Branch>')],
            'plugin':    [_('Plugin(_P)'), None, 1, 1, ('/plugin', None, None, 0, '<Branch>')]
            }
        self.__pixmap = None
        self.__ghost_list = None
        self.__shell_list = None
        self.__balloon_list = None
        self.__plugin_list = None
        self.__fontcolor = "#000000"
        self.__sidebar_width = 0
        menu_items = (
            self.__menu_list['portal'][-1],
            self.__menu_list['recommend'][-1],
            ('/Separator', None, None, 0, '<Separator>'),
            self.__menu_list['file'][-1],
            self.__menu_list['update'][-1],
            self.__menu_list['vanish'][-1],
            self.__menu_list['reload'][-1],
            self.__menu_list['settings'][-1],
            self.__menu_list['usage'][-1],
            self.__menu_list['version'][-1],
            self.__menu_list['exit'][-1],
            self.__menu_list['change'][-1],
            self.__menu_list['summon'][-1],
            self.__menu_list['shell'][-1],
            self.__menu_list['balloon'][-1],
            ('/Separator', None, None, 0, '<Separator>'),
            self.__menu_list['scaling'][-1],
            self.__menu_list['costume'][-1],
            self.__menu_list['stick'][-1],
            self.__menu_list['wait'][-1],
            ('/Separator', None, None, 0, '<Separator>'),
            self.__menu_list['nekodorif'][-1],
            self.__menu_list['kinoko'][-1],
            self.__menu_list['plugin'][-1],
            )
        self.__item_factory = gtk.ItemFactory(gtk.Menu, '<main>', accelgroup)
        self.__item_factory.create_items(menu_items)
        self.__popup_menu = self.__item_factory.get_widget('<main>')
        for key in self.__menu_list.keys():
            path = self.__menu_list[key][-1][0]
            self.__menu_list[key][1] = self.__item_factory.get_item(path)
        self.__create_scale_menu()
        self.__create_speed_menu()
    def __create_scale_menu(self):
        menu = gtk.Menu()
        item = gtk.CheckMenuItem(_('Scale Balloon'))
        if self.__ghost.get_balloon_scalling():
            item.set_active(gtk.TRUE)
        else:
            item.set_active(gtk.FALSE)
        item.connect("activate", self.toggle_balloon_scalling)
        menu.add(item)
        item.show()
        item = gtk.SeparatorMenuItem()
        menu.add(item)
        item.show()
        group = None
        for label, value in ninix.surface.range_scale:
            item = group = gtk.RadioMenuItem(group, label)
            item.set_name("popup menu item")
            item.connect("activate", self.select_surface_scale, value)
            if value == self.__ghost.get_surface_scale():
                item.set_active(gtk.TRUE)
            item.show()
            menu.append(item)
        menuitem = self.__menu_list['scaling'][1]
        menuitem.set_submenu(menu)
        self.__menu_list['scaling'][3] = 1 # changed
    def __create_speed_menu(self):
        menu = gtk.Menu()
        group = None
        for label, value in ninix.sakura.range_script_speed:
            item = group = gtk.RadioMenuItem(group, unicode(label, 'utf-8'))
            item.set_name("popup menu item")
            item.connect("activate", self.select_script_speed, value)
            if value == self.__surface.sakura.get_script_speed():
                item.set_active(gtk.TRUE)
            item.show()
            menu.append(item)
        menuitem = self.__menu_list['wait'][1]
        menuitem.set_submenu(menu)
        self.__menu_list['wait'][3] = 1 # changed
    def select_surface_scale(self, widget, value):
        ##print "select_surface_scale", value
        self.__ghost.set_surface_scale(value)
    def select_script_speed(self, widget, value):
        ##print "select_script_speed", value
        self.__surface.sakura.set_script_speed(value)
    def toggle_balloon_scalling(self, widget, value=None):
        if widget.get_active():
            balloon_scalling = 1
        else:
            balloon_scalling = 0
        self.__ghost.set_balloon_scalling(balloon_scalling)
    def set_fontcolor(self, r, g ,b):
        self.__fontcolor = "#%02x%02x%02x" % (r, g, b)
    def set_pixmap(self, path_background, path_sidebar):
        pixbuf = None
        if path_background and os.path.exists(path_background):
            try:
                pixbuf = ninix.pix.create_pixbuf_from_file(path_background, 'png')
            except:
                pixbuf = None
        if pixbuf == None:
            self.__pixmap = None
            return
        pixbuf_sidebar = None
        if path_sidebar and os.path.exists(path_sidebar):
            try:
                pixbuf_sidebar = ninix.pix.create_pixbuf_from_file(path_sidebar, 'png')
            except:
                pixbuf_sidebar = None
        if pixbuf_sidebar:
            width = pixbuf.get_width()
            height = pixbuf.get_height()
            sidebar_width = pixbuf_sidebar.get_width()
            sidebar_height = pixbuf_sidebar.get_height()
            new_pixbuf = gtk.gdk.Pixbuf(
                pixbuf.get_colorspace(), gtk.FALSE,
                pixbuf.get_bits_per_sample(),
                width + sidebar_width, max(height, sidebar_height))
            pixbuf.composite(
                new_pixbuf,
                0, 0, new_pixbuf.get_width(), new_pixbuf.get_height(),
                width - 1, height - 1,
                1.0, 1.0, gtk.gdk.INTERP_BILINEAR, 255)
            pixbuf_sidebar.composite(
                new_pixbuf,
                0, 0, sidebar_width, sidebar_height,
                0, 0,
                1.0, 1.0, gtk.gdk.INTERP_BILINEAR, 255)
            pixbuf.composite(
                new_pixbuf,
                sidebar_width, 0, width, height,
                0, 0,
                1.0, 1.0, gtk.gdk.INTERP_BILINEAR, 255)
            pixbuf = new_pixbuf
            self.__sidebar_width = sidebar_width
        else:
            self.__sidebar_width = 0
        pixmap, mask = pixbuf.render_pixmap_and_mask(255)
        self.__pixmap = pixmap
    def __set_mayuna_menu(self, side):
        if self.__mayuna_menu[side] != None:
            menuitem = self.__menu_list['costume'][1]
            menuitem.set_submenu(self.__mayuna_menu[side])
            self.__set_visible('costume', 1)
        else:
            self.__set_visible('costume', 0)
    def create_mayuna_menu(self, mayuna_menu):
        self.__mayuna_menu = []
        for side, index in [('sakura', 0), ('kero', 1)]:
            if mayuna_menu[side]:
                self.__mayuna_menu.append(gtk.Menu())
                item = gtk.TearoffMenuItem()
                item.show()
                if self.__pixmap:
                    item.window.set_back_pixmap(self.__pixmap, gtk.FALSE)
                self.__mayuna_menu[index].append(item)
                for j in range(len(mayuna_menu[side])):
                    key, name, state = mayuna_menu[side][j]
                    if key != '-':
                        item = gtk.CheckMenuItem(name)
                        item.set_name("popup menu item")
                        if state:
                            item.set_active(gtk.TRUE)
                        else:
                            item.set_active(gtk.FALSE)
                        item.connect("activate", self.__surface.toggle_bind, (index, key))
                    else:
                        item = gtk.SeparatorMenuItem()
                    item.show()
                    self.__mayuna_menu[index].append(item)
            else:
                self.__mayuna_menu.append(None)
    __re_shortcut = re.compile(r'&(?=[\x21-\x7e])')
    def __modify_shortcut(self, caption):
        return self.__re_shortcut.sub('_', caption)
    __re_accel = re.compile(r'\(_.\)|_')
    def __cut_accel(self, caption):
        return self.__re_accel.sub('', caption)
    __ui = {'update': ([["updatebuttoncaption", "updatebutton.caption"], ["updatebuttoncaption", "updatebutton.caption"]], "(_U)", [[],[]]),
            'vanish': ([["vanishbuttoncaption", "vanishbutton.caption"], ["vanishbuttoncaption", "vanishbutton.caption"]], "(_F)", [["vanishbuttonvisible", "vanishbutton.visible"], ["vanishbuttonvisible", "vanishbutton.visible"]]),
            'portal': ([["sakura.portalbuttoncaption", "portalrootbutton.caption"], []], "(_P)", [[], None]),
            'recommend': ([["sakura.recommendbuttoncaption", "recommendrootbutton.caption"], ["kero.recommendbuttoncaption"]], "(_R)", [[], []]),
           }
    def __update_ui(self, side):
        for key in self.__ui.keys():
            if not self.__menu_list.has_key(key):
                continue
            list = self.__ui[key][0][side]
            if list: # caption
                for name in list:
                    caption = self.__ghost.getstring(name)
                    if caption:
                        break
                if caption:
                    caption = self.__modify_shortcut(caption)
                    if caption == self.__cut_accel(caption):
                        caption += self.__ui[key][1]
                    self.__set_caption(key, caption)
            list = self.__ui[key][2][side]
            if list: # visible
                for name in list:
                    visible = self.__ghost.getstring(name)
                    if visible is not None:
                        break
                if visible == "0":
                    self.__set_visible(key, 0)
                else:
                    self.__set_visible(key, 1)
            elif visible is None:
                self.__set_visible(key, 0)
    def popup(self, button, side): ## FIXME
        self.__update_ui(side)
        if side == 0:
            portal = self.__ghost.getstring("sakura.portalsites")
        else:
            portal = None
        self.__set_portal_menu(side, portal)
        recommend = self.__ghost.getstring(["sakura.recommendsites", "kero.recommendsites"][side])
        self.__set_recommend_menu(side, recommend)
        self.__set_ghost_menu()
        self.__set_shell_menu()
        self.__set_balloon_menu()
        self.__set_plugin_menu()
        self.__set_mayuna_menu(side)
        self.__set_nekodorif_menu()
        self.__set_kinoko_menu()
        for key in self.__menu_list.keys():
            text = self.__menu_list[key][0]
            item = self.__menu_list[key][1]
            visible = self.__menu_list[key][2]
            changed = self.__menu_list[key][3]
            if item and changed:
                self.__menu_list[key][3] = 0 # changed
                if visible:
                    label = item.get_children()[0]
                    markup = '   <span foreground="%s">%s</span>' % (self.__fontcolor, text)
                    label.set_markup_with_mnemonic(markup)
                    item.show()
                    if self.__pixmap: ## FIXME
                        if key in ['file', 'update']:
                            item = self.__menu_list[key][1]
                            item.realize()
                            item.window.set_back_pixmap(self.__pixmap, gtk.FALSE)
                        if key in ['portal', 'recommend', 'costume', 'scaling', 'wait', 'change', 'summon', 'shell', 'balloon', 'plugin', 'nekodorif', 'kinoko']:
                            item_list = self.__menu_list[key][1].get_submenu().get_children()
                            if len(item_list) > 0: ## FIXME: nested submenu
                                item = item_list[0]
                                item.realize()
                                item.window.set_back_pixmap(self.__pixmap, gtk.FALSE)
                            for item in item_list:
                                # reset label
                                children = item.get_children()
                                if len(children) > 0:
                                    label = children[0]
                                    text = label.get_text()
                                    markup = '   <span foreground="%s">%s</span>' % (self.__fontcolor, text)
                                    label.set_markup_with_mnemonic(markup)
                else:
                    item.hide()
        try:
            self.__popup_menu.popup(None, None, None, button, gtk.get_current_event_time())
        except: # pygtk <= 2.0.0
            self.__popup_menu.popup(None, None, None, button, 0)
    def __set_caption(self, name, caption):
        assert self.__menu_list.has_key(name)
        assert type(caption) in [types.StringType, types.UnicodeType]
        if self.__menu_list[name][2] == caption:
            return
        self.__menu_list[name][0] = caption
        self.__menu_list[name][3] = 1 # changed
    def __set_visible(self, name, visible):
        assert self.__menu_list.has_key(name)
        assert visible in [0, 1]
        if self.__menu_list[name][2] == visible:
            return
        self.__menu_list[name][2] = visible
        self.__menu_list[name][3] = 1 # changed
    def __set_portal_menu(self, side, portal): ## FIXME
        if side == 1:
            self.__set_visible('portal', 0)
        else:
            if portal:
                menu = gtk.Menu()
                list = portal.split(chr(2))
                for site in list:
                    entry = site.split(chr(1))
                    if len(entry) < 1:
                        continue
                    title = entry[0]
                    if title == '-':
                        item = gtk.SeparatorMenuItem()
                    else:
                        if len(entry) < 2:
                            continue
                        url = entry[1]
                        if len(entry) > 2:
                            bannar = entry[2] ## FIXME
                        else:
                            bannar = None
                        item = gtk.MenuItem(title)
                        item.connect("activate", self.__ghost.launch_browser, (title, url))
                    menu.add(item)
                    item.show()
                self.__menu_list['portal'][1].set_submenu(menu)
                menu.show()
                self.__set_visible('portal', 1)
            else:
                self.__set_visible('portal', 0)
    def __set_recommend_menu(self, side, recommend): ## FIXME
        if recommend:
            menu = gtk.Menu()
            list = recommend.split(chr(2))
            for site in list:
                entry = site.split(chr(1))
                if len(entry) < 1:
                    continue
                title = entry[0]
                if title == '-':
                    item = gtk.SeparatorMenuItem()
                else:
                    if len(entry) < 2:
                        continue
                    url = entry[1]
                    if len(entry) > 2:
                        bannar = entry[2] ## FIXME
                    else:
                        bannar = None
                    item = gtk.MenuItem(title)
                    item.connect("activate", self.__ghost.launch_browser, (title, url))
                menu.add(item)
                item.show()
            self.__menu_list['recommend'][1].set_submenu(menu)
            menu.show()
            self.__set_visible('recommend', 1)
        else:
            self.__set_visible('recommend', 0)
    def __update_ghost_menu(self, list): ## FIXME
        ##if self.__ghost_list == None:
        ##    self.__create_ghost_menu(list)
        ##else:
        ghost_menu = gtk.Menu()
        for i in range(len(list)):
            if list[i] == None:
                continue
            name = list[i]['name']
            icon_path = list[i]['icon']
            pixbuf = None
            if icon_path != None:
                try:
                    pixbuf = ninix.pix.create_pixbuf_from_file(icon_path, 'ico')
                except:
                    pixbuf = None
            shell_list = list[i]['shell']
            if pixbuf is not None:
                pixbuf = pixbuf.scale_simple(16, 16, gtk.gdk.INTERP_BILINEAR)
                image = gtk.Image()
                image.set_from_pixbuf(pixbuf)
                image.show()
                item = gtk.ImageMenuItem(name)
                item.set_image(image)
            else:
                item = gtk.MenuItem(name)
            item.set_name("popup menu item")
            item.show()
            ghost_menu.append(item)
            if len(shell_list) <= 1:
                shell_name, value = shell_list[0]
                item.connect("activate", self.__ghost.app.select_sakura, (self.__ghost, value)) ## FIXME
                if list[i]['instance'].is_running():
                    item.set_sensitive(gtk.FALSE)
            else:
                submenu = gtk.Menu()
                submenu.set_name("popup menu")
                item.set_submenu(submenu)
                for shell_name, value in shell_list:
                    item = gtk.MenuItem(shell_name)
                    item.set_name("popup menu item")
                    item.connect("activate", self.__ghost.app.select_sakura, (self.__ghost, value)) ## FIXME
                    item.show()
                    if list[i]['instance'].is_running():
                        if self.__ghost.current[1] != value[1]:
                            item.set_sensitive(gtk.FALSE)
                        elif self.__ghost.current[2] == value[2]:
                            item.set_sensitive(gtk.FALSE)
                    submenu.append(item)
        menuitem = self.__menu_list['change'][1]
        menuitem.set_submenu(ghost_menu)
        self.__menu_list['change'][3] = 1
        self.__ghost_list = list
    def __set_ghost_menu(self): ## FIXME
        list = self.__ghost.app.get_ghost_list()
        self.__update_ghost_menu(list)
        self.__update_summon_menu(list)
    def __update_summon_menu(self, list): ## FIXME
        ##if self.__ghost_list == None:
        ##    self.__create_ghost_menu(list)
        ##else:
        ghost_menu = gtk.Menu()
        for i in range(len(list)):
            if list[i] == None:
                continue
            name = list[i]['name']
            icon_path = list[i]['icon']
            pixbuf = None
            if icon_path != None:
                try:
                    pixbuf = ninix.pix.create_pixbuf_from_file(icon_path, 'ico')
                except:
                    pixbuf = None
            shell_list = list[i]['shell']
            if pixbuf is not None:
                pixbuf = pixbuf.scale_simple(16, 16, gtk.gdk.INTERP_BILINEAR)
                image = gtk.Image()
                image.set_from_pixbuf(pixbuf)
                image.show()
                item = gtk.ImageMenuItem(name)
                item.set_image(image)
            else:
                item = gtk.MenuItem(name)
            item.set_name("popup menu item")
            item.show()
            ghost_menu.append(item)
            if len(shell_list) <= 1:
                shell_name, value = shell_list[0]
                item.connect("activate", self.__ghost.app.start_sakura_cb, value) ## FIXME
                if list[i]['instance'].is_running():
                    item.set_sensitive(gtk.FALSE)
            else:
                submenu = gtk.Menu()
                submenu.set_name("popup menu")
                item.set_submenu(submenu)
                for shell_name, value in shell_list:
                    item = gtk.MenuItem(shell_name)
                    item.set_name("popup menu item")
                    item.connect("activate", self.__ghost.app.start_sakura_cb, value) ## FIXME
                    item.show()
                    if list[i]['instance'].is_running():
                        item.set_sensitive(gtk.FALSE)
                    submenu.append(item)
        menuitem = self.__menu_list['summon'][1]
        menuitem.set_submenu(ghost_menu)
        self.__menu_list['summon'][3] = 1
        self.__ghost_list = list
    def __update_shell_menu(self, list): ## FIXME
        ##if self.__shell_list == None:
        ##    self.__create_shell_menu(list)
        ##else:
        shell_menu = gtk.Menu()
        for i in range(len(list)):
            name = list[i]['name']
            item = gtk.MenuItem(name)
            item.set_name("popup menu item")
            item.show()
            shell_menu.append(item)
            shell_list = list[i]['shell']
            if len(shell_list) <= 1:
                shell_name, value = shell_list[0]
                item.connect("activate", self.__ghost.app.select_sakura, (self.__ghost, value)) ## FIXME
                ##if working:
                ##    item.set_sensitive(gtk.FALSE)
            else:
                submenu = gtk.Menu()
                submenu.set_name("popup menu")
                item.set_submenu(submenu)
                for shell_name, value in shell_list:
                    item = gtk.MenuItem(shell_name)
                    item.set_name("popup menu item")
                    item.connect("activate", self.__ghost.app.select_sakura, (self.__ghost, value)) ## FIXME
                    item.show()
                    ##if working: ## FIXME
                    ##    item.set_sensitive(gtk.FALSE)
                    submenu.append(item)
        menuitem = self.__menu_list['shell'][1]
        menuitem.set_submenu(shell_menu)
        self.__menu_list['shell'][3] = 1
        self.__shell_list = list
    def __set_shell_menu(self): ## FIXME
        list = self.__ghost.app.get_shell_list()
        self.__update_shell_menu(list)
    def __update_balloon_menu(self, list): ## FIXME
        ##if self.__balloon_list == None:
        ##    self.__create_balloon_menu(list)
        ##else:
        balloon_menu = gtk.Menu()
        group = None
        name_own = self.__ghost.get_own_balloon_name()
        if name_own != None:
            item = group = gtk.RadioMenuItem(group, name_own)
            item.set_name("popup menu item")
            hid = item.connect("activate", self.__ghost.select_balloon, name_own)
            item.show()
            balloon_menu.append(item)
        item = gtk.SeparatorMenuItem()
        balloon_menu.add(item)
        item.show()
        for i in range(len(list)):
            name = list[i]['name']
            item = group = gtk.RadioMenuItem(group, name)
            item.set_name("popup menu item")
            hid = item.connect("activate", self.__ghost.select_balloon, name)
            item.show()
            balloon_menu.append(item)
            if name == self.__ghost.get_current_balloon_name():
                item.handler_block(hid)
                item.activate()
                item.handler_unblock(hid)
        menuitem = self.__menu_list['balloon'][1]
        menuitem.set_submenu(balloon_menu)
        self.__menu_list['balloon'][3] = 1
        self.__balloon_list = list
    def __set_balloon_menu(self): ## FIXME
        list = self.__ghost.app.get_balloon_list()
        self.__update_balloon_menu(list)
    def __update_plugin_menu(self, list): ## FIXME
        ##if self.__plugin_list == None:
        ##    self.__create_plugin_menu(list)
        ##else:
        plugin_menu = gtk.Menu()
        for i in range(len(list)):
            name = list[i]['name']
            item = gtk.MenuItem(name)
            item.set_name("popup menu item")
            item.show()
            plugin_menu.append(item)
            item_list = list[i]['items']
            if len(item_list) <= 1:
                label, value = item_list[0]
                item.connect("activate", self.__ghost.app.select_plugin, value) ## FIXME
                ##if working:
                ##    item.set_sensitive(gtk.FALSE)
            else:
                submenu = gtk.Menu()
                submenu.set_name("popup menu")
                item.set_submenu(submenu)
                for label, value in item_list:
                    item = gtk.MenuItem(label)
                    item.set_name("popup menu item")
                    item.connect("activate", self.__ghost.app.select_plugin, value) ## FIXME
                    item.show()
                    ##if working:
                    ##    item.set_sensitive(gtk.FALSE)
                    submenu.append(item)
        menuitem = self.__menu_list['plugin'][1]
        menuitem.set_submenu(plugin_menu)
        self.__menu_list['plugin'][3] = 1
        self.__plugin_list = list
    def __set_plugin_menu(self): ## FIXME
        list = self.__ghost.app.get_plugin_list()
        self.__update_plugin_menu(list)
    def __update_nekodorif_menu(self, list): ## FIXME
        ##if self.__nekodorif_list == None:
        ##    self.__create_nekodorif_menu(list)
        ##else:
        nekodorif_menu = gtk.Menu()
        for i in range(len(list)):
            name = list[i]['name']
            item = gtk.MenuItem(name)
            item.set_name("popup menu item")
            item.show()
            nekodorif_menu.append(item)
            item.connect("activate", self.__ghost.app.select_nekodorif, (self.__ghost, list[i]['dir'])) ## FIXME
            ##if working:
            ##    item.set_sensitive(gtk.FALSE)
        menuitem = self.__menu_list['nekodorif'][1]
        menuitem.set_submenu(nekodorif_menu)
        self.__menu_list['nekodorif'][3] = 1
        self.__nekodorif_list = list
    def __set_nekodorif_menu(self): ## FIXME
        list = self.__ghost.app.get_nekodorif_list()
        self.__update_nekodorif_menu(list)
    def __update_kinoko_menu(self, list): ## FIXME
        ##if self.__kinoko_list == None:
        ##    self.__create_kinoko_menu(list)
        ##else:
        kinoko_menu = gtk.Menu()
        for i in range(len(list)):
            name = list[i]['title']
            item = gtk.MenuItem(name)
            item.set_name("popup menu item")
            item.show()
            kinoko_menu.append(item)
            item.connect("activate", self.__ghost.app.select_kinoko, (self.__ghost, list[i])) ## FIXME
            ##if working:
            ##    item.set_sensitive(gtk.FALSE)
        menuitem = self.__menu_list['kinoko'][1]
        menuitem.set_submenu(kinoko_menu)
        self.__menu_list['kinoko'][3] = 1
        self.__kinoko_list = list
    def __set_kinoko_menu(self): ## FIXME
        list = self.__ghost.app.get_kinoko_list()
        self.__update_kinoko_menu(list)
    def get_stick(self):
        item = self.__menu_list['stick'][1]
        if item and item.get_active():
            return 1
        else:
            return 0

def test():
    pass

if __name__ == "__main__":
    test()
